`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// LPF 1 tap
// 2020.12.8 Arata Kawamura
//////////////////////////////////////////////////////////////////////////////////

module axis_proc #(
    parameter SWITCH_WIDTH = 4, // Number of Slide Switch
    parameter DATA_WIDTH = 24
) (
    input wire clk,
    input wire [SWITCH_WIDTH-1:0] sw,
    
    //AXIS SLAVE INTERFACE
    input  wire [DATA_WIDTH-1:0] s_axis_data,
    input  wire s_axis_valid,
    output reg  s_axis_ready = 1'b1,
    input  wire s_axis_last,
    
    // AXIS MASTER INTERFACE
    output reg [DATA_WIDTH-1:0] m_axis_data = 1'b0,
    output reg m_axis_valid = 1'b0,
    input  wire m_axis_ready,
    output reg m_axis_last = 1'b0
);

    wire m_select      = m_axis_last;
    wire m_new_word    = (m_axis_valid == 1'b1 && m_axis_ready == 1'b1) ? 1'b1 : 1'b0;
    wire m_new_packet  = (m_new_word == 1'b1 && m_axis_last == 1'b1) ? 1'b1 : 1'b0;
    
    wire s_select      = s_axis_last;
    wire s_new_word    = (s_axis_valid == 1'b1 && s_axis_ready == 1'b1) ? 1'b1 : 1'b0;
    wire s_new_packet  = (s_new_word == 1'b1 && s_axis_last == 1'b1) ? 1'b1 : 1'b0;
    reg s_new_packet_r = 1'b0;
    

    //**********************************************************//
    //*        Setting for Fixed Point Calculation             *//
    //*                                                        *//
    //*        [integer(8bits)].[decimal(40bits)]              *//
    //*   ---> SignBit(8bits)_input(24bits)_0(16bits)          *//
    //*                                                        *//
    //**********************************************************//
    parameter TBL = 48;                  // Total Bit Length including Sign Bit
    parameter DPL = 40;                  // Decimal Part Length

    parameter D_Len = 1;                // Memory length for delayed signals

    //**********************************************************//
    //*                Declaration of Signals                   *//
    //**********************************************************//
    reg [TBL-1:0] xin  [1:0];            // input signal
    reg [TBL-1:0] yout [1:0];            // output signal
    reg [TBL-1:0] x_L  [0:D_Len];        // left  delayed input  signal
    reg [TBL-1:0] x_R  [0:D_Len];        // right delayed input  signal
    reg [TBL-1:0] y_L  [0:D_Len];        // left  delayed output signal
    reg [TBL-1:0] y_R  [0:D_Len];        // right delayed output signal

    //**********************************************************//
    //*                Declaration of Paramters                *//
    //**********************************************************//
    reg signed [TBL-1:0] a  =   (5)*(2**DPL)/10;    // a = 0.5
    
    /* Examples of decalaration 
    reg signed [TBL-1:0] x;
    reg signed [TBL-1:0] u  =   (5)*(2**DPL);    // u = 5
    reg signed [TBL-1:0] h  =   (1)*(2**DPL)/10; // h = 0.1
    reg signed [TBL-1:0] a  =  (-1)*(2**DPL)/10; // a = -0.1
    reg signed [TBL-1:0] th =  (80)*(2**DPL);    // th = 80
    reg signed [TBL-1:0] b  = (-80)*(2**DPL);    // b = -80
    */
        
    always@(posedge clk) begin
        if (s_new_word == 1'b1)          // get 24bits input data and sign extend --> 48bits
            xin[s_select] <= { {8{s_axis_data[DATA_WIDTH-1]}}, s_axis_data, {16{1'b0}} };
        else if (s_new_packet_r == 1'b1) begin
        
            //**********************************************************//
            //*                                                        *//
            //*                  Signal Processing                     *//
            //*                                                        *//
            //*              Sampling Frequency 44.1kHz                *//
            //*                    2 channels                          *//
            //**********************************************************//
            x_L[0]  = $signed( xin[0] );             // left  input signal (48bits)
            x_R[0]  = $signed( xin[1] );             // right input signal (48bits)

            y_L[0]  = $signed( FPM(a,x_L[0]) ) - $signed( FPM(a,x_L[1]) );    // left   output
            y_R[0]  = $signed( FPM(a,x_R[0]) ) - $signed( FPM(a,x_R[1]) );    // right  output

            x_L[1]  <= x_L[0];                        // shift left  input
            x_R[1]  <= x_R[0];                        // shift right input
            
            //**********************************************************//
            //*               End of Signal Processing                 *//
            //**********************************************************//
            //**************************************//
            //*                                    *//
            //*  sw[0]=0: Through                  *//
            //*  sw[0]=1: Processing               *//
            //*                                    *//
            //**************************************//
            yout[0] = (sw[0])? $signed(y_L[0]) : $signed(xin[0]); // left  channel output
            yout[1] = (sw[0])? $signed(y_R[0]) : $signed(xin[1]); // right channel output
        end
    end
        
    always@(posedge clk) begin
        s_new_packet_r <= s_new_packet;
    end
    
    always@(posedge clk)
        if (s_new_packet_r == 1'b1)
            m_axis_valid <= 1'b1;
        else if (m_new_packet == 1'b1)
            m_axis_valid <= 1'b0;
            
    always@(posedge clk)
        if (m_new_packet == 1'b1)
            m_axis_last <= 1'b0;
        else if (m_new_word == 1'b1)
            m_axis_last <= 1'b1;
            
    //************************************************************//
    //*  D/A <-- [24bits] <-- yout is [8bits].[24bits][16bits]   *//
    //***********************************************************s*//
    always@(m_axis_valid, yout[0], yout[1], m_select)
        if (m_axis_valid == 1'b1)
            m_axis_data = yout[m_select][39:16];
        else
            m_axis_data = 'b0;
            
    always@(posedge clk)
        if (s_new_packet == 1'b1)
            s_axis_ready <= 1'b0;
        else if (m_new_packet == 1'b1)
            s_axis_ready <= 1'b1;

    /////////////////////////////
    // Multiplication Function //
    /////////////////////////////
    function [TBL-1:0] FPM(input [TBL-1:0] IN1, input [TBL-1:0] IN2);
    reg signed [2*TBL-1:0] MUL;
    begin
        MUL = $signed(IN1) * $signed(IN2);
        MUL = MUL / (2**DPL);
        FPM = $signed({MUL[2*TBL-1], MUL[TBL-2:0]});
    end
    endfunction
    
    ///////////////////////
    // Division Function //
    ///////////////////////
    function [TBL-1:0] FPD(input [TBL-1:0] IN1, input [TBL-1:0] IN2);
    reg signed [2*TBL-1:0] DIV;
    begin
        DIV = $signed(IN1) * $signed(2**DPL);
        DIV = DIV / $signed(IN2);
        FPD = $signed({DIV[2*TBL-1], DIV[TBL-2:0]});
    end
    endfunction
            
endmodule
